<?php

/* goal.php - SynApp2 database navigation/state manager
**
   +----------------------------------------------------------------------+
   | SynApp2 Version 1                                                    |
   +----------------------------------------------------------------------+
   | Copyright (c) 2007 - 2009 Richard Howell. All rights reserved.       |
   +----------------------------------------------------------------------+
   | This source file is subject to version 1.01 of the SynApp2 license,  |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.synapp2.org/license/1_01.txt                              |
   | If you did not receive a copy of the SynApp2 license and are unable  |
   | to obtain it through the world-wide-web, please send a note to       |
   | license@synapp2.org so we can mail you a copy immediately.           |
   +----------------------------------------------------------------------+
   | Authors: Richard Howell                                              |
   |                                                                      |
   +----------------------------------------------------------------------+
**
** http://www.synapp2.org
**
** $Id: goal.php,v 1.2 2009/01/05 02:39:54 richard Exp $
*/

class goal
{
    var $m_action = null;
    var $m_custom = null;
    var $m_schema = null;

    var $m_is_first = false;

    var $m_elements = null;

    var $m_pid = null;
    var $m_guide = null;

    var $m_goal_state = null;

    var $m_container = null;
    var $m_forms = null;
    var $m_stat = null;
    var $m_followup = null;

    function goal(&$action, &$custom, &$schema, $pid, $objective, $guidance = null)
    {
        $this->m_action = &$action;
        $this->m_custom = &$custom;
        $this->m_schema = &$schema;

        $this->m_pid = $pid;

        if (!empty($objective))
        {
            $this->build($objective);

//add_debug_msg($this->m_goal_state, 'initial state');

            if (!empty($guidance))
            {
                list($guide_pid, $guide_objective) = explode(';', $guidance);

                assert('!empty($guide_pid)');
                assert('!empty($guide_objective)');

                if (!empty($guide_pid) && !empty($guide_objective))
                {
                    $this->m_guide = new goal($action, $custom, $schema, $guide_pid, $guide_objective);
                }
            }
        }
    }

    ////

    function build($objective)
    {
        if (!$this->fetch_goal_state())
        {
            $this->m_is_first = true;

            $tokens = explode(',', $objective); // basis_table,basis_col,tier_table,basis_col,tier_table...

            if (is_array($tokens))
            {
                $this->m_goal_state->set_basis_table(current($tokens));

                if (count($tokens) == 1)
                {
                    // set up for single-node (standalone) table navigation

                    $table_name = $this->m_goal_state->get_basis_table();
                    $col_name = $this->m_custom->get_pk_name($table_name);

                    $this->m_goal_state->m_inodes[] = 0;
                    $this->m_goal_state->m_goal_nodes[][] = array(TABLE_NAME=>$table_name,
                                                                  COL_NAME=>$col_name,
                                                                  JOIN_TABLE=>$table_name,
                                                                  JOIN_COL=>$col_name,
                                                                  CONSTRAINT=>null,
                                                                  OFFSET=>0);
                }
                else
                {
                    // set up for multi-node table navigation

                    $paths = $this->m_schema->get_paths($this->m_goal_state->get_basis_table());

                    while ($basis_col = next($tokens))
                    {
                        if ($tier_table = next($tokens))
                        {
                            foreach ($paths as $path)
                            {
                                $goals = array();

                                foreach ($path[PATH] as $path_node)
                                {
                                    if (!empty($goals) || ($path_node[TABLE_NAME] == $this->m_goal_state->get_basis_table() &&
                                                           $path_node[COL_NAME] == $basis_col))
                                    {
                                        $path_node[CONSTRAINT] = null;
                                        $path_node[OFFSET] = 0;

                                        $goals[] = $path_node;

                                        if ($path_node[JOIN_TABLE] == $tier_table)
                                        {
                                            $path_node[TABLE_NAME] = $path_node[JOIN_TABLE];
                                            $path_node[COL_NAME] = $path_node[JOIN_COL];

                                            $goals[] = $path_node;

                                            $this->m_goal_state->m_inodes[] = 0;
                                            $this->m_goal_state->m_goal_nodes[] = array_reverse($goals);
                                            break 2;
                                        }
                                    }
                                }
                            }
                        }
                        else
                        {
                            break;
                        }
                    }
                }
            }
        }

        if ($this->m_action->get_nav() < count($this->m_goal_state->m_goal_nodes))
        {
            $this->m_goal_state->m_ipath = $this->m_action->get_nav(); // TODO: experimental
        }
    }

    ////

    function on_get_norm()
    {
        if (!$this->m_action->get_pk())
        {
            // search

            $table = $this->m_goal_state->get_table();
            $nav = $this->get_nav();

            if ($this->m_goal_state->is_objective_satisfied())
            {
                $this->m_container = "select:{$nav}:{$table}";
                $this->m_forms = "select:{$nav}:{$table}";
                $this->m_stat = "stat:{$nav}:{$table}";
            }
            else if ($this->m_is_first)
            {
                $this->m_is_first = false;
                $this->m_container = "search:{$nav}:{$table}";
                $this->m_forms = "search:{$nav}:{$table}";
            }
            else
            {
                $this->m_container = "select:{$nav}:{$table}";
                $this->m_forms = "search:{$nav}:{$table},select:{$nav}:{$table}";
                $this->m_stat = "stat:{$nav}:{$table}";
            }
        }
        else
        {
            $this->m_goal_state->set_offset($this->m_action->get_limit_offset());

            if (!$this->m_goal_state->is_objective_satisfied(true))
            {
                // select

                $this->m_goal_state->step();
                $this->m_goal_state->set_constraint($this->m_action->get_pk());

                $parent_table = $this->m_goal_state->get_table(true);
                $table = $this->m_goal_state->get_table();
                $nav = $this->get_nav();

                if (!$this->m_goal_state->is_objective_satisfied())
                {
                    $this->m_forms = "display:{$nav}:{$parent_table},search:{$nav}:{$table}";
                }
                else
                {
                    $this->m_forms = "display:{$nav}:{$parent_table}";
                }

                $this->m_container = "display:{$nav}:{$parent_table}";
                $this->m_followup = "select:{$nav}:{$table}";
            }
            else
            {
                // edit/delete

                $table = $this->m_goal_state->get_table();
                $nav = $this->get_nav();

                $this->m_container = "input:{$nav}:{$table}";
                $this->m_forms = "input:{$nav}:{$table}";
            }
        }
    }

    function on_get_init()
    {
        // add

        $this->m_goal_state->set_offset($this->m_action->get_limit_offset());

        $table = $this->m_goal_state->get_table();
        $nav = $this->get_nav();

        $this->m_container = "input:{$nav}:{$table}";
        $this->m_forms = "input:{$nav}:{$table}";
    }

    function on_put_norm()
    {
        // insert/update/delete

        $this->on_cancel();
    }

    function on_cancel()
    {
        $table = $this->m_goal_state->get_table();
        $nav = $this->get_nav();

        if ($this->m_goal_state->is_standalone())
        {
            $this->m_container = "search:{$nav}:{$table}";
            $this->m_forms = "search:{$nav}:{$table},select:{$nav}:{$table}";
        }
        else
        {
            $this->m_container = "select:{$nav}:{$table}";
            $this->m_forms = "select:{$nav}:{$table}";
        }

        $this->m_followup = "select:{$nav}:{$table}";
    }

    function on_close()
    {
        $is_first = $this->m_goal_state->m_inodes[$this->m_goal_state->m_ipath] == 0;

        $table = $this->m_goal_state->get_table();
        $nav = $this->get_nav();

        $rollup = "rollup:{$nav}:{$table}";

        $this->m_goal_state->set_constraint(null);
        $this->m_goal_state->set_offset(0);
        $this->m_goal_state->step(true);

        $table = $this->m_goal_state->get_table();
        $nav = $this->get_nav();

        if ($is_first)
        {
            $this->m_container = "search:{$nav}:{$table}";
            $this->m_forms = "{$rollup},search:{$nav}:{$table},select:{$nav}:{$table},stat:{$nav}:{$table}";
        }
        else
        {
            $this->m_container = "select:{$nav}:{$table}";
            $this->m_forms = "{$rollup},search:{$nav}:{$table},select:{$nav}:{$table},stat:{$nav}:{$table}";
            $this->m_followup = "select:{$nav}:{$table}";
        }
    }

    ////

    function process()
    {
        assert('empty($this->m_forms)');

        if (!empty($this->m_goal_state) &&
            empty($this->m_forms) && // check for multiple invocations
           ($method = 'on_' . $this->m_action->get_request_mode()) &&
            method_exists($this, $method))
        {
            $this->m_container = '';
            $this->m_forms = '';
            $this->m_stat = '';
            $this->m_followup = '';

            $this->$method();

//add_debug_msg($this->m_goal_state, 'resulting state');

            $this->save_goal_state();
        }
    }

    ////

    function get_offset()
    {
        return $this->m_goal_state ? $this->m_goal_state->get_offset() : 0 ;
    }

    function get_action()
    {
        $this->m_elements = null;

        if (!empty($this->m_forms))
        {
            if (!empty($this->m_container))
            {
                $this->m_elements[] = '<goal_container>';
                $this->m_elements[] = $this->m_container;
                $this->m_elements[] = '</goal_container>';
            }

            $this->m_elements[] = '<goal_action>';
            $this->m_elements[] = $this->m_forms;
            $this->m_elements[] = '</goal_action>';

            if (!empty($this->m_stat))
            {
                $this->m_elements[] = '<goal_stat>';
                $this->m_elements[] = $this->m_stat;
                $this->m_elements[] = '</goal_stat>';
            }

            if (!empty($this->m_followup))
            {
                $this->m_elements[] = '<goal_followup>';
                $this->m_elements[] = $this->m_followup;
                $this->m_elements[] = '</goal_followup>';
            }
        }

        return $this->m_elements;
    }

    function get_fk_constraint()
    {
        if ($this->m_guide)
        {
            return $this->m_guide->get_fk_constraint();
        }

        $fk_constraint = array('n' => '', 'v' => 0);

        // TODO: this method will need some study and work to handle multi-path goals

        if (!empty($this->m_goal_state->m_goal_nodes) && $this->m_goal_state->get_constraint() && ($goal_node = $this->m_goal_state->get_goal_node()))
        {
            $fk_constraint['n'] = $goal_node[TABLE_NAME] . '.' . $goal_node[COL_NAME];
            $fk_constraint['v'] = $goal_node[CONSTRAINT];
        }

        return $fk_constraint;
    }

    function get_fk_fwd($alternate_basis_table = null)
    {
        if ($this->m_guide)
        {
            return $this->m_guide->get_fk_fwd($this->m_goal_state->get_basis_table());
        }

        $fk_fwd = null;

        $ipath = 0; // TODO: this method will need some study and work to handle multi-path goals

        if ($this->m_goal_state && !empty($this->m_goal_state->m_goal_nodes[$ipath]))
        {
            $basis_table = $alternate_basis_table ? $alternate_basis_table : $this->m_goal_state->get_basis_table() ;

            foreach ($this->m_goal_state->m_goal_nodes[$ipath] as $i => $goal_node)
            {
                if ($goal_node[TABLE_NAME] == $basis_table && !empty($goal_node[CONSTRAINT]))
                {
                    $fk_fwd[$goal_node[COL_NAME]] = $goal_node[CONSTRAINT];
                    break;
                }
            }
        }

        return $fk_fwd;
    }

    ////

    function get_nav()
    {
        return $this->m_action->get_nav();
    }

    function map_independent_keys()
    {
        $col_vars = array();

        $keys = $this->m_action->get_keys();

        if (!empty($keys))
        {
            foreach ($keys as $nav_key)
            {
                list($nav, $v) = explode(',', $nav_key);

                if (isset($this->m_goal_state->m_goal_nodes[$nav][0]))
                {
                    $last_node = count($this->m_goal_state->m_goal_nodes[$nav]) - 1;

                    $n = $this->m_goal_state->m_goal_nodes[$nav][$last_node][COL_NAME];

                    $col_vars[$n][] = $v; // the $nav value has now been decoded into a basis_table (fk) col_name
                }
            }

            if (!empty($col_vars))
            {
                // this is admittedly goofy, but it transparently inserts the col_vars
                // in the same format as the query object expects them to be
                //
                foreach ($col_vars as $n => $a)
                {
                    $v = count($a) == 1 ? $a[0] : $a ; // TODO: TEMP: convert from multi-value format

                    assert('empty($this->m_action->m_col_vars[$n])');
                    $this->m_action->m_col_vars[$n] = array('n' => $n, 'v' => $v); // TODO: HACK: KLUGE: need insert method
//                    $this->m_action->m_col_vars[$n] = &$_POST[$n]; // TODO:
                    $this->m_action->m_col_expands[$n] = $v; // TODO: HACK: KLUGE: need insert method
                }
//add_debug_msg($this->m_action->get_col_vars(), '$this->m_action->get_col_vars()');
            }
        }
    }

    ////

    function fetch_goal_state()
    {
        if ($serialized_state = get_session_var('state_' . $this->m_pid))
        {
            $this->m_goal_state = unserialize($serialized_state);
        }
        else
        {
            $this->m_goal_state = new goal_state();
        }

        return !empty($serialized_state);
    }

    function save_goal_state()
    {
        if ($this->m_goal_state->is_dirty(true))
        {
            $serialized_state = serialize($this->m_goal_state);
            set_session_var('state_' . $this->m_pid, $serialized_state);
        }
    }
}

////////

class goal_state
{
    var $m_basis_table = '';
    var $m_ipath = 0;
    var $m_inodes = array();
    var $m_goal_nodes = array();

    var $m_dirty = false;

    function goal_state()
    {
    }

    ////

    function get_goal_node()
    {
        return !empty($this->m_goal_nodes) ? $this->m_goal_nodes[$this->m_ipath][$this->m_inodes[$this->m_ipath]] : null ;
    }

    ////

    function get_table($get_parent = false)
    {
        $goal_node = $this->get_goal_node();

        return !empty($goal_node) ? ($get_parent ? $goal_node[JOIN_TABLE] : $goal_node[TABLE_NAME]) : null ;
    }

    function get_constraint()
    {
        return !empty($this->m_goal_nodes) ? $this->m_goal_nodes[$this->m_ipath][$this->m_inodes[$this->m_ipath]][CONSTRAINT] : null ;
    }

    function get_offset()
    {
        return !empty($this->m_goal_nodes) ? $this->m_goal_nodes[$this->m_ipath][$this->m_inodes[$this->m_ipath]][OFFSET] : 0 ;
    }

    function get_basis_table()
    {
        return $this->m_basis_table;
    }

    ////

    function is_standalone()
    {
        return count($this->m_goal_nodes) == 1 && count($this->m_goal_nodes[$this->m_ipath]) == 1;
    }

    function is_objective_satisfied($pk_flag = false, $independent = true)
    {
        return ($this->get_constraint() || $pk_flag) &&
               ($independent || $this->m_ipath == count($this->m_goal_nodes) - 1) &&
               $this->m_inodes[$this->m_ipath] == count($this->m_goal_nodes[$this->m_ipath]) - 1;
    }

    ////

    function is_dirty($reset_flag = false)
    {
        if (($is_dirty = $this->m_dirty !== false) && $reset_flag)
        {
            $this->m_dirty = false;
        }

        return $is_dirty;
    }

    ////

    function set_basis_table($table_name)
    {
        $this->m_basis_table = $table_name;
        $this->m_dirty = true;
    }

    function set_constraint($value)
    {
        $this->m_goal_nodes[$this->m_ipath][$this->m_inodes[$this->m_ipath]][CONSTRAINT] = $value;
        $this->m_dirty = true;
    }

    function set_offset($value)
    {
        $this->m_goal_nodes[$this->m_ipath][$this->m_inodes[$this->m_ipath]][OFFSET] = $value;
        $this->m_dirty = true;
    }

    function step($backward = false, $independent = true)
    {
        if (!$this->step_node_index($backward) && !$independent)
        {
            $this->step_path_index($backward);
        }
    }

    function step_node_index($backward = false)
    {
        $delta = $backward ? -1 : 1 ;

        $i = $this->m_inodes[$this->m_ipath] + $delta;

        if ($i >= 0 && $i < count($this->m_goal_nodes[$this->m_ipath]))
        {
            $this->m_inodes[$this->m_ipath] = $i;
            $this->m_dirty = true;
        }

        return $i == $this->m_inodes[$this->m_ipath];
    }

    function step_path_index($backward = false)
    {
        $delta = $backward ? -1 : 1 ;

        $i = $this->m_ipath + $delta;

        if ($i >= 0 && $i < count($this->m_goal_nodes))
        {
            $this->m_ipath = $i;
            $this->m_dirty = true;
        }

        return $i == $this->m_ipath;
    }
}

////

?>
