<?php

/* validate.php - SynApp2 data validation and reporting
**
   +----------------------------------------------------------------------+
   | SynApp2 Version 1                                                    |
   +----------------------------------------------------------------------+
   | Copyright (c) 2007 - 2009 Richard Howell. All rights reserved.       |
   +----------------------------------------------------------------------+
   | This source file is subject to version 1.01 of the SynApp2 license,  |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.synapp2.org/license/1_01.txt                              |
   | If you did not receive a copy of the SynApp2 license and are unable  |
   | to obtain it through the world-wide-web, please send a note to       |
   | license@synapp2.org so we can mail you a copy immediately.           |
   +----------------------------------------------------------------------+
   | Authors: Richard Howell                                              |
   |                                                                      |
   +----------------------------------------------------------------------+
**
** http://www.synapp2.org
**
** $Id: validate.php,v 1.5 2009/06/12 21:35:34 richard Exp $
*/

function validate_columns(&$action, &$dbx, &$custom, &$schema)
{
    $feedback = null;

    if ($action->is_insert() || $action->is_update())
    {
        $table_name = $custom->get_table();
        $cols_info = $schema->get_table_cols_info($table_name);
        $fk_fwd = $action->get_fk_fwd();

        foreach ($cols_info as $col_info)
        {
            $col_name = $col_info['Field'];
            $display_name = $schema->get_col_display_name($col_info);

            $col_type_detail = $schema->get_col_type_detail($col_info);

            if ($schema->is_pk($table_name, $col_name))
            {
                // ok - do nothing // TODO: ASSUME: all pk's are auto_increment
            }
            else
            {
                $is_fk = $schema->is_fk($table_name, $col_name);
                $is_blind = $schema->is_blind($table_name, $col_name);

                $col_var = $action->get_col_var($col_name);
                $col_var_is_present = $col_var !== null;
                $col_var_value = $col_var_is_present ? (is_array($col_var) ? trim($col_var[0]) : trim($col_var)) : '' ;
                $col_var_value_is_empty = strlen($col_var_value) < 1;

                if ($is_fk)
                {
                    $display_name .= ' (FK)';

                    if (isset($fk_fwd[$col_name]))
                    {
                        $col_var_value = $fk_fwd[$col_name];
                        $col_var_value_is_empty = strlen($col_var_value) < 1;

                        //assert('$action->is_insert()'); // TODO: TBD: is $fk_fwd relevant for $action->is_update()??
                        assert('!$col_var_is_present'); // multiple fkey: $col_name
                        assert('!empty($col_var_value);'); // missing fkey: $col_name
                    }
                }

                if ($is_fk && $action->is_insert() && $col_var_value_is_empty && !is_valid_empty($col_info))
                {
                    $feedback[$col_name] = array('status'=>"$display_name is required!");
                }
                else if ($is_fk && $action->is_update() && !$col_var_is_present)
                {
                    // ok - do nothing;
                }
                else if ($is_blind && (!$col_var_is_present || $col_var_value_is_empty))
                {
                    // ok - do nothing;
                }
                else if ($col_var_value_is_empty && !is_valid_empty($col_info))
                {
                    $feedback[$col_name] = array('status'=>"$display_name is required, please enter a value");
                }
                else if ($col_var_value_is_empty)
                {
                    // ok - do nothing;
                }
                else if ($type_msg = validate_type($dbx, $col_info, $col_type_detail, $col_var_value))
                {
                    $feedback[$col_name] = array('status'=>"$display_name must be a valid $type_msg");
                }
                else if ($length_msg = validate_length($col_info, $col_type_detail, $col_var_value))
                {
                    $feedback[$col_name] = array('status'=>"$display_name may not be longer than $length_msg");
                }
                else if ($range_msg = validate_range($col_info, $col_type_detail, $col_var_value))
                {
                    $feedback[$col_name] = array('status'=>"$display_name must be $range_msg");
                }
                else if (!validate_duplicate($col_info, $col_type_detail, $col_var_value))
                {
                    $feedback[$col_name] = array('status'=>"$display_name must be unique");
                }
            }
        }
    }
    else if ($action->is_delete())
    {
        $join_table = $custom->get_table();
        $pk_name = $custom->get_pk_name($join_table, true);
        $pk = $action->get_pk();

        $path_nodes = $schema->get_children($join_table);

        if (!empty($path_nodes))
        {
            foreach ($path_nodes as $path_node)
            {
                $table_name = $path_node[TABLE_NAME];

                $col_names = $custom->get_child_join_cols($join_table, $table_name);

                if (!empty($col_names))
                {
                    foreach ($col_names as $col_name)
                    {
                        $join_col = $custom->get_parent_join_col($join_table, $table_name, $col_name);

                        $qty = $dbx->get_query_value("select count(*) from $table_name, $join_table where $pk_name = '$pk' and $join_table.$join_col = $table_name.$col_name");

                        if ($qty > 0)
                        {
                            $s = $qty > 1 ? 's' : '' ;
                            $feedback[$join_table]  = array('status' => "delete_denied - $qty dependent record{$s} in table: $table_name");
                        }
                    }
                }
            }
        }
    }

    return $feedback;
}

function validate_unique(&$action, &$dbx, &$custom, &$schema)
{
    $feedback = null;

    if ($error_dup = $dbx->get_error_dup())
    {
        $msg_dup = "duplicate_denied - Your entry would result a duplicate record."; // TODO: get custom msg_dup for $qid
        $feedback["Table: {$custom->get_table()}"] = array('status'=>$msg_dup);
    }

    return $feedback;
}

////

function validate_type(&$dbx, &$col_info, &$col_type_detail, $col_value)
{
    $type_msg = '';

    $col_type = strtoupper($col_type_detail['type']);
    $is_type_unsigned = strpos(strtoupper($col_type_detail['flags']), 'UNSIGNED') !== false;

    if (strpos($col_type, 'CHAR') !== false ||
        strpos($col_type, 'TEXT') !== false)
    {
        // ok - do nothing
    }
    else if (strpos($col_type, 'ENUM') !== false ||
             strpos($col_type, 'SET') !== false)
    {
        if (!empty($col_type_detail['values']) &&
            is_array($col_type_detail['values']) &&
            !in_array($col_value, $col_type_detail['values']))
        {
            $type_msg = "'$col_type' member";
        }
    }
    else if ($col_type == 'DATE')
    {
        $ymd = $dbx->parse_date_value($col_value);

        if (!is_valid_date($ymd['y'], $ymd['m'], $ymd['d']))
        {
            $type_msg = "'$col_type' (" . $dbx->get_datepicker_format() . ")";
        }
    }
    else if ($col_type == 'TIME')
    {
        list($h, $m, $s) = sscanf($col_value, "%d:%d:%d");

        if (!ereg('^[0-9]{2}:[0-9]{2}:[0-9]{2}$', $col_value) ||
            !is_valid_time($h, $m, $s))
        {
            $type_msg = "'$col_type' (hh:mm:ss)";
        }
    }
    else if ($col_type == 'DATETIME' ||
             $col_type == 'TIMESTAMP')
    {
        if ($col_type != 'TIMESTAMP' || $col_value != '0000-00-00 00:00:00')
        {
            list ($y, $m, $d, $h, $m, $s) = sscanf($col_value, "%d-%d-%d %d:%d:%d");

            if (!ereg('^[0-9]{4}-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2}$', $col_value) ||
                !is_valid_date($y, $m, $d, $col_type == 'TIMESTAMP') ||
                !is_valid_time($h, $m, $s))
            {
                $type_msg = "'$col_type' (yyyy-mm-dd hh:mm:ss)";
            }
        }
    }
    else if ($col_type == 'YEAR')
    {
        if (!ereg('^[0-9]{4}$', $col_value) ||
            !is_valid_year($col_value))
        {
            $type_msg = "'$col_type' (yyyy)";
        }
    }
    else if (strpos($col_type, 'INT') !== false ||
                    $col_type == 'BIT')
    {
        $regex = '^' . ($is_type_unsigned ? '' : '[-]?') . '[0-9]+$';

        if (!ereg($regex, $col_value))
        {
            $type_msg = "'" . ($is_type_unsigned ? 'UNSIGNED ' : '') . "$col_type'";
        }
    }
    else if (strpos($col_type, 'FLOAT') !== false ||
             strpos($col_type, 'DOUBLE') !== false ||
             strpos($col_type, 'DECIMAL') !== false ||
             strpos($col_type, 'NUMERIC') !== false ||
             strpos($col_type, 'REAL') !== false)
    {
        if (!is_numeric($col_value) || ($is_type_unsigned && 0.0 > $col_value))
        {
            $type_msg = "'" . ($is_type_unsigned ? 'UNSIGNED ' : '') . "$col_type'";
        }
    }
    else if (strpos($col_type, 'BLOB') !== false ||
             strpos($col_type, 'BINARY') !== false)
    {
        // ok - do nothing
    }
    else
    {
        // unknown type // TODO: TBD: message??
    }

    return $type_msg;
}

function validate_length(&$col_info, &$col_type_detail, $col_value)
{
    $maxlength = '';

    $col_type = strtoupper($col_type_detail['type']);

    if (strpos($col_type, 'CHAR') !== false)
    {
        $length = strlen($col_value);

        if ($length > $col_type_detail['maxlength'])
        {
            $maxlength = $col_type_detail['maxlength'] . " (currently $length)";
        }
    }
    else
    {
        // ok - do nothing
    }

    return $maxlength;
}

////

function is_valid_empty(&$col_info)
{
    return eregi('^Y', $col_info['Null']) !== false;
}

function is_valid_date($y, $m, $d, $is_timestamp = false)
{
    return $y >= ($is_timestamp ? 1970 : 1000) &&
           $y <= ($is_timestamp ? 2037 : 9999) &&
           checkdate($m, $d, $y); // TODO: ASSUME: MySQL limits
}

function is_valid_year($y)
{
    return $y >= 1901 && $y <= 2155; // TODO: ASSUME: MySQL limits (approx)
}

function is_valid_time($h, $m, $s)
{
    return $h >= 0 && $h <= 23 &&
           $m >= 0 && $m <= 59 &&
           $s >= 0 && $s <= 59;
}

////

function get_value_range(&$col_info, &$col_type_detail)
{
    $value_range = null;

    if (!empty($col_info['Comment']))
    {
        $csv = explode(',', $col_info['Comment']);

        for ($i = 0; $i < count($csv); $i++)
        {
            if (eregi('([a-z]*) *= *([-]?[0-9]+(\.[0-9]*)?)', $csv[$i], $matches))
            {
                $n = strtoupper($matches[1]);
                $v = $matches[2];

                switch ($n)
                {
                case 'MIN':
                case 'MAX':
                    $value_range[$n] = $v;
                    break;
                default:
                    // unknown
                    break;
                }
            }
        }
    }

    return $value_range;
}

function validate_range(&$col_info, &$col_type_detail, $value)
{
    $range_msg = '';

    $value_range = get_value_range($col_info, $col_type_detail);

    // TODO: add support for !is_numeric($value)

    if (isset($value_range['MIN']) && $value < $value_range['MIN'])
    {
        $range_msg = "greater than or equal to '{$value_range['MIN']}'";
    }
    else if (isset($value_range['MAX']) && $value > $value_range['MAX'])
    {
        $range_msg = "less than or equal to '{$value_range['MAX']}'";
    }

    return $range_msg;
}

function validate_duplicate(&$col_info, &$col_type_detail, $value)
{
    $result = true;

    // TODO: not implemented

    return $result;
}

?>
